try:
    import configparser
except ImportError:
    import ConfigParser as configparser
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import logging
import os
import sys
from os import path
import re
from ceph_medic import terminal, metadata

logger = logging.getLogger(__name__)

medic_conf_template = """
#
# ceph-medic configuration file
#

[global]
# Overrides for some of ceph-medic's global flags, like verbosity or cluster
# name Cluster Name. Ceph defaults this to 'ceph' when not specified, and
# ceph-medic will try to infer this if the configuration is not found on the
# first '.conf' file in /etc/ceph/
--cluster = ceph
# Use a specific location for the Ansible inventory file, it defaults to look
# into the current working directory and /etc/ansible (in that order) unless
# specified here or directly at the command line
#--inventory
#
# The logging verbosity, this will affect both the terminal and file logging
# levels
--verbosity = info
#
# Should always be an absolute path, although '.' is allowed to log from
# wherever the CLI is executed from (current working directory)
--log-path = .

[check]
# Overrides for some of ceph-medic's check flags, like what errors or warnings
# to ignore. By default nothing is ignored (everything is reported). Ignores
# should be comma separated into their respective codes.
# --ignore = ECOMM101,ECOMM102

"""


class _TrimIndentFile(object):
    """
    This is used to take a file-like object and removes any
    leading tabs from each line when it's read. This is important
    because some ceph configuration files include tabs which break
    ConfigParser.
    """
    def __init__(self, fp):
        self.fp = fp

    def readline(self):
        line = self.fp.readline()
        return line.lstrip(' \t')

    def __iter__(self):
        return iter(self.readline, '')


def location():
    """
    Find and return the location of the ceph-medic configuration file. If this
    file does not exist, create one in a default location.
    """
    return _locate_or_create()


def get_host_file(_path=None):
    """
    If ``_path`` is passed in, return that (makes it easier for the caller if
    it can pass a ``None``) otherwise look into the most common locations for
    a host file and if found, return that.
    """
    if _path:
        return _path
    # if no path is passed onto us try and look in the cwd for a hosts file
    if os.path.isfile('hosts'):
        logger.info(
            'found and loaded the hosts file from the current working directory: %s',
            os.getcwd()
        )
        return path.abspath('hosts')

    # if that is not the case, try for /etc/ansible/hosts
    if path.isfile('/etc/ansible/hosts'):
        return '/etc/ansible/hosts'
    logger.warning('unable to find an Ansible hosts file to work with')
    logger.warning('tried locations: %s, %s', os.getcwd(), '/etc/ansible/hosts')


def load_hosts(_path=None):
    """
    Very similar to a plain configuration load, but do not fail if a hosts file
    is not loaded. Try the hardest to load one, and return an empty dictionary
    if nothing is found.
    """
    if _path and not os.path.exists(_path):
        terminal.error("the given inventory path does not exist: %s" % _path)
        sys.exit()
    _path = _path or get_host_file()
    return AnsibleInventoryParser(_path)


def parse_hosts(conf):
    """
    From a parsed hosts file, loop through all the nodes and get the actual
    hosts from each line
    """
    parsed_hosts = {}
    if not conf:
        return parsed_hosts
    for section in conf.sections:
        parsed_hosts[section] = []


def load_string(conf_as_string):
    """
    An easy way to get the INI configuration as a string, but return a Conf
    object after writing it to a StringIO (file-like) object.
    """
    file_obj = StringIO()
    file_obj.write(conf_as_string)
    file_obj.seek(0)
    return load(file=_TrimIndentFile(file_obj))


def load(path=None, file=None):
    parser = Conf()
    try:
        if file:
            parser.readfp(file)
        elif path and os.path.exists(path):
            parser.read(path)
        else:
            parser.read(location())
        return parser
    except configparser.ParsingError as error:
        terminal.error('Unable to read configuration file')
        terminal.error(str(error))
        logger.exception('Unable to parse INI-style file')


def _locate_or_create():
    home_config = path.expanduser('~/.cephmedic.conf')
    # With order of importance
    locations = [
        path.join(os.getcwd(), 'cephmedic.conf'),
        home_config,
    ]

    for location in locations:
        if path.exists(location):
            logger.debug('found configuration file at: %s' % location)
            return location
    logger.info('could not find configuration file, will create one in $HOME')
    create_stub(home_config)
    return home_config


def create_stub(_path=None):
    _path = _path or path.expanduser('~/.cephmedic.conf')
    logger.debug('creating new configuration file: %s' % _path)
    with open(_path, 'w') as cm_conf:
        cm_conf.write(medic_conf_template)


def get_overrides(_conf=None):
    """
    Read the configuration file and look for ceph-medic sections and flags to
    set the defaults, these are later checked by the main method to see if any
    need to be overridden *again* if specified on the CLI directly.
    """
    # Get the subcommand name to avoid overwritting values from other
    # subcommands that are not going to be used
    conf = _conf or load()
    conf_arguments = {}
    for section_name in conf.sections():
        if section_name == 'global':
            for k, v in conf.items(section_name):
                conf_arguments[k] = v
        else:
            conf_arguments[section_name] = {}
            for k, v in conf.items(section_name):
                conf_arguments[section_name][k] = v

    return conf_arguments


class Conf(configparser.SafeConfigParser):
    """
    Subclasses from SafeConfigParser to give a few helpers for the ceph-medic
    configuration. Specifically, it addresses the need to work with Ansible
    host sections and some custom values that might be comma-separated
    """

    def get_safe(self, section, key, default=None):
        """
        Attempt to get a configuration value from a certain section
        in a ``cfg`` object but returning None if not found. Avoids the need
        to be doing try/except {ConfigParser Exceptions} every time.
        """
        try:
            return self.get(section, key)
        except (configparser.NoSectionError, configparser.NoOptionError):
            return default

    def get_list(self, section, key):
        """
        Assumes that the value for a given key is going to be a list separated
        by commas. It gets rid of trailing comments.  If just one item is
        present it returns a list with a single item, if no key is found an
        empty list is returned.
        """
        value = self.get_safe(section, key, [])
        if value == []:
            return value

        # strip comments
        value = re.split(r'\s+#', value)[0]

        # split on commas
        value = value.split(',')

        # strip spaces
        return [x.strip() for x in value]

    def optionxform(self, s):
        """
        Ceph configuration allows both these forms (considering them equal)::

            some key option = 1
            some_key_option = 1

        This method ensures that regardless of any of the formats, all options
        will be treated as using the underscores. This is also how ceph-deploy
        reads Ceph configuration values
        """
        s = s.replace('_', ' ')
        s = '_'.join(s.split())
        return s


class AnsibleInventoryParser(object):
    """
    Based roughly on Ansible's ``InventoryParser``, but without capturing
    specifics on host or group vars because ceph-medic only cares about the
    groups and the hosts, and ConfigParser spits bullets with Ansible's lax INI
    format.

    It borrows the regexes from ``InventoryParser`` to match groups and sections, with the ability
    of expanding deeply nested grouping to understand what hosts belong to an actual group.

    Given this hosts file::

        [osds:children]
        usa

        [atlanta]
        osd0

        [florida]
        mon0

        [usa:children]
        atlanta
        florida

    Will make this parser add ``osd0`` as part of ``osds`` so that ceph-medic
    understands what the actual role of that node is. This extra bit of
    expanding and re-doing what Ansible could do for us is done so that we
    aren't requiring all of Ansible just to be able to load a hosts file.

    .. todo:: A try/except import for ``InventoryParser`` should be attempted
              so that we can just use Ansible if present, but fallback to this
              implementation if it isn't.
    """

    def __init__(self, filename=None):
        self.filename = filename
        self.hosts = {}
        self.patterns = {}
        self.groups = {}
        self.nodes = {}

        # Read the hosts file, stripping comments and empty lines
        # which are not needed for parsing
        contents = []

        if filename:
            with open(filename, 'r') as fh:
                for line in fh.readlines():
                    if line.startswith(u';'):
                        continue
                    if line.startswith(u'#'):
                        continue
                    if line:
                        contents.append(line.strip())

        self._parse(contents)

    def _parse(self, lines):
        """
        Read through all the ``lines`` of the hosts file, sorting the groups,
        hosts, and finally the nodes.
        """

        self._compile_patterns()

        # Define `ungrouped` first since hosts might not belong anywhere (why
        # would you do this though). Ansible supports this so we are trying to
        # get it supported here as well
        groupname = 'ungrouped'
        state = 'hosts'

        for line in lines:
            host_item = {'host': None, 'group': None}

            m = self.patterns['section'].match(line)
            if m:
                (groupname, state) = m.groups()
                state = state or 'hosts'
                self.groups[groupname] = []
                self.nodes[groupname] = []
                continue
            if state == 'vars':
                continue
            if state == 'hosts':
                host = self._parse_host_definition(line)
                if host:
                    host_item['host'] = host
                    host_item['group'] = groupname
                    self.nodes[groupname].append(host_item)
            if state == 'children':
                child = self._parse_group_name(line)
                if groupname not in self.groups:
                    self.groups[groupname] = []
                if child and child not in self.groups[groupname]:
                    self.groups[groupname].append(child)
                host_item['parent'] = groupname

            host_group = host_item.get('group')
            if host_group:
                if host_group not in self.hosts:
                    self.hosts[host_group] = [host_item]
                else:
                    self.hosts[host_group].append(host_item)
        self.expand_nodes()
        self.filter_groups()

    def filter_groups(self):
        """
        Removes any groups from self.nodes that do not exist in
        ceph_medic.metadata. We want to do this because we don't care
        about nodes or groups in an inventory that are not part of
        the ceph cluster.
        """
        groups = [key for key in self.nodes.keys()]
        for group in groups:
            if group not in metadata:
                del self.nodes[group]

    def expand_nested_group(self, parent_group, group):
        """
        For a ``group`` list, try to identify if any members are parents of
        other groups, if that is the case, keep recursing until the bottom of
        the membership is identified so that all the hosts can be added to
        ``self.nodes``
        """
        for g in group:
            if self.groups.get(g):
                # this group exists as a parent to some other group so recurse
                # and try to get to the bottom of the nesting
                self.expand_nested_group(parent_group, self.groups.get(g))
            if self.hosts.get(g):
                # This group is the direct parent of host(s), so extend the
                # members of that list
                self.nodes[parent_group].extend(self.hosts.get(g))

    def expand_nodes(self):
        """
        After all groups and hosts have been parsed and their memberships identified, it will then
        look at the group membership of all groups and determine which is the primary parent of the
        tree, so that we can eventually look at ``nodes['osds']`` and retrieve all the hosts that we
        need to know about.
        """
        # now reverse lookup, identifying if groups have other parents
        # so that it can populate the ``nodes`` attribute
        for parent_group, child_groups in self.groups.items():
            if child_groups:
                # pre-populate the value so that we can extend this later
                # without overriding anything that might be already there
                self.nodes[parent_group] = self.nodes.get(parent_group, [])
            for group in child_groups:

                if self.groups.get(group): # means we have a nested parent/child here
                    self.expand_nested_group(parent_group, self.groups[group])

                # this group has direct nodes already so extend
                # the nodes present for that one group
                if self.hosts.get(group):
                    self.nodes[parent_group].extend(self.hosts.get(group))

    def _parse_host_definition(self, line):
        """
        Identify a host from a single line. It supports hosts on the following formats::

            [osds]
            osd0
            osd1:6789 docker=True
            osd3 docker=True

        If the line is empty it will return a ``None`` which is caught by the caller.
        """
        if not line:
            return

        # split on whitespace
        host = line.split()[0]

        # split on colon
        host = host.split(':')[0]
        return host

    def _parse_group_name(self, line):
        """
        Try to match a groupname (which is usually in the format of
        Takes a single line and tries to parse it as a group name. Returns the
        group name if successful, or raises an error.
        """

        m = self.patterns['groupname'].match(line)
        if m:
            return m.group(1)

    def _compile_patterns(self):
        """
        Compiles the regexes based on the original implementation of the parser (``InventoryParser``)
        and sets the to ``self.patterns``
        """
        self.patterns['section'] = re.compile(
            r'''^\[
                    ([^:\]\s]+)             # group name
                    (?::(\w+))?             # optional tag name :
                \]
                \s*                         # ignore trailing whitespace
                (?:\#.*)?                   # and/or a comment till the
                $                           # end of the line
            ''', re.X
        )

        self.patterns['groupname'] = re.compile(
            r'''^
                ([^:\]\s]+)
                \s*                         # ignore trailing whitespace
                (?:\#.*)?                   # and/or a comment till the
                $                           # end of the line
            ''', re.X
        )



import sys
import ceph_medic
from ceph_medic.connection import get_connection
import remoto
import json
from tambo import Transport


def generate_inventory(inventory, to_stdout=False, tmp_dir=None):
    """
    Generates a host file to use with an ansible-playbook call.

    The first argument is a dictionary mapping that contains the group name as
    the key and a list of hostnames as values

    For example:

        {'mons': ['mon.host'], 'osds': ['osd1.host', 'osd1.host']}
    """
    result = []
    for section, hosts in inventory.items():
        group_name = section
        result.append("[{0}]".format(group_name))
        if not isinstance(hosts, list):
            hosts = [hosts]
        result.extend(hosts)
    result_str = "\n".join(result) + "\n"
    # if not None the NamedTemporaryFile will be created in the given directory
    if to_stdout:
        print result_str
        return
    with open('hosts_file', 'w') as hosts_file:
        hosts_file.write(result_str)


def get_mon_report(conn):
    out, err, code = remoto.process.check(
        conn,
        [
            'ceph',
            'report'
        ],
    )

    if code > 0:
        for line in err:
            print line

    try:
        return json.loads(b''.join(out).decode('utf-8'))
    except ValueError:
        return {}


class Generate(object):
    help = "Create a hosts file (Ansible compatible) from the information on a running Ceph cluster"
    long_help = """
Create a hosts file (Ansible compatible) from the information on a running Ceph
cluster.

Usage:

    ceph-medic generate [/path/to/ceph.conf]
    ceph-medic generate [MONITOR HOST]

Loaded Config Path: {config_path}

    """

    def __init__(self, argv=None, parse=True):
        self.argv = argv or sys.argv

    def _help(self):
        skip_internal = ['__file__', 'config_path', 'verbosity']
        node_section = []
        for daemon, node in ceph_medic.config['nodes'].items():
            if daemon in skip_internal or not node:
                continue
            header = "\n* %s:\n" % daemon
            body = '\n'.join(["    %s" % n for n in ceph_medic.config['nodes'][daemon].keys()])
            node_section.append(header+body+'\n')
        return self.long_help.format(
            config_path=ceph_medic.config['config_path']
        )

    def main(self):
        options = ['--stdout']
        parser = Transport(
            self.argv, options=options,
            check_version=False
        )
        parser.catch_help = self._help()

        parser.parse_args()

        if len(self.argv) == 1:
            raise SystemExit("A monitor hostname or a ceph.conf file is required as an argument")

        node = self.argv[-1]
        inventory = {}

        with get_connection(node) as conn:
            report = get_mon_report(conn)
            try:
                mons = report['monmap']['mons']
            except KeyError:
                raise SystemExit(report)
            inventory['mons'] = [i['name'] for i in mons]
            osds = report['osd_metadata']
            inventory['osds'] = [i['hostname'] for i in osds]

        if not inventory:
            raise SystemExit('no hosts where found from remote monitor node: %s' % node)

        generate_inventory(inventory, to_stdout=parser.get('--stdout'))
        conn.exit()
        return



import os
import pytest
from textwrap import dedent
from ceph_medic.util import configuration


def make_hosts_file(filename, contents=None):
    contents = contents or "[mons]\nmon0\n[osds]\nosd0\n"
    with open(filename, 'w') as f:
        f.write(contents)


class TestFlatInventory(object):

    def test_parses_both_sections(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'hosts')
        make_hosts_file(filename)
        result = configuration.AnsibleInventoryParser(filename)
        assert sorted(result.nodes.keys()) == sorted(['mons', 'osds'])

    def test_populates_hosts(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'hosts')
        make_hosts_file(filename)
        result = configuration.AnsibleInventoryParser(filename).nodes
        assert result['mons'][0]['host'] == 'mon0'
        assert result['osds'][0]['host'] == 'osd0'

    def test_hosts_do_not_get_mixed(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'hosts')
        make_hosts_file(filename)
        result = configuration.AnsibleInventoryParser(filename).nodes
        assert len(result['mons']) == 1
        assert len(result['osds']) == 1

    def test_ignores_unknown_groups(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'hosts')
        contents = """
        [mons]
        mon0

        [test]
        node1
        """
        make_hosts_file(filename, contents)
        result = configuration.AnsibleInventoryParser(filename).nodes
        assert 'test' not in result

    def test_hosts_file_does_not_exist(self):
        with pytest.raises(SystemExit):
            configuration.load_hosts(_path="/fake/path")


class TestNestedInventory(object):

    def test_nested_one_level(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'hosts')
        contents = """
        [mons:children]
        atlanta

        [atlanta]
        mon0
        """
        make_hosts_file(filename, contents)
        result = configuration.AnsibleInventoryParser(filename).nodes
        assert result['mons'][0]['host'] == 'mon0'

    def test_nested_one_level_populates_other_groups(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'hosts')
        contents = """
        [mons:children]
        atlanta

        [atlanta]
        mon0
        """
        make_hosts_file(filename, contents)
        result = configuration.AnsibleInventoryParser(filename).nodes
        assert result['mons'][0]['host'] == 'mon0'

    def test_nested_levels_populates(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'hosts')
        contents = """
        [mons:children]
        us

        [atlanta]
        mon0

        [us:children]
        atlanta
        """
        make_hosts_file(filename, contents)
        result = configuration.AnsibleInventoryParser(filename).nodes
        assert result['mons'][0]['host'] == 'mon0'


class TestLoadString(object):

    def test_loads_valid_ceph_key(self):
        contents = dedent("""
        [global]
        cluster = ceph
        """)
        conf = configuration.load_string(contents)
        assert conf.get_safe('global', 'cluster') == 'ceph'

    def test_loads_key_with_spaces_converted(self):
        contents = dedent("""
        [global]
        some key here = ceph
        """)
        conf = configuration.load_string(contents)
        assert conf.get_safe('global', 'some_key_here') == 'ceph'



from ceph_medic.checks import common
from ceph_medic import metadata


class TestGetFsid(object):

    def setup(self):
        metadata['cluster_name'] = 'ceph'

    def make_metadata(self, contents=None):
        contents = contents or ''
        data = {'paths': {'/etc/ceph':{'files':{'/etc/ceph/ceph.conf':{'contents': contents}}}}}
        data['cluster_name'] = 'ceph'
        return data

    def test_fails_to_find_an_fsid(self):
        data = self.make_metadata("[global]\nkey=value\n\n[mdss]\ndisabled=true\n")
        fsid = common.get_fsid(data)
        assert fsid == ''

    def test_empty_conf_returns_empty_string(self):
        data = self.make_metadata()
        fsid = common.get_fsid(data)
        assert fsid == ''

    def test_find_an_actual_fsid(self):
        data = self.make_metadata("[global]\nfsid = 1234-lkjh\n\n[mdss]\ndisabled=true\n")
        fsid = common.get_fsid(data)
        assert fsid == '1234-lkjh'

    def test_spaces_on_fsid_are_trimmed(self):
        data = self.make_metadata("[global]\nfsid = 1234-lkjh   \n\n[mdss]\ndisabled=true\n")
        fsid = common.get_fsid(data)
        assert fsid == '1234-lkjh'


class TestCephVersionParity(object):

    def setup(self):
        metadata['cluster_name'] = 'ceph'

    def test_finds_a_mismatch_of_versions(self, make_nodes, make_data):
        metadata['nodes'] = make_nodes(mons=['node1', 'node2'])
        node1_data = make_data()
        metadata['mons']['node1'] = node1_data
        metadata['mons']['node2'] = make_data({'ceph': {'version': '13'}})
        result = common.check_ceph_version_parity('node1', node1_data)
        assert 'Ceph version "12.2.1" is different' in str(result)

    def test_versions_have_parity(self, make_nodes, make_data):
        metadata['nodes'] = make_nodes(mons=['node1', 'node2'])
        node1_data = make_data()
        metadata['mons']['node1'] = node1_data
        metadata['mons']['node2'] = make_data()
        result = common.check_ceph_version_parity('node1', node1_data)
        assert result is None


class TestCephSocketAndInstalledVersionParity(object):

    def setup(self):
        metadata['cluster_name'] = 'ceph'

    def test_finds_a_mismatch_of_versions(self, make_nodes, make_data):
        metadata['nodes'] = make_nodes(mons=['node1'])
        node1_data = make_data(
            {'ceph': {
                "sockets": {
                    "/var/run/ceph/osd.asok": {"version": "13.2.0"},
                },
                "installed": True,
                "version": "12.2.1",
            }}
        )
        metadata['mons']['node1'] = node1_data
        result = common.check_ceph_socket_and_installed_version_parity('node1', node1_data)
        assert 'Ceph version "12.2.1" is different' in str(result)

    def test_versions_have_parity(self, make_nodes, make_data):
        metadata['nodes'] = make_nodes(mons=['node1'])
        node1_data = make_data(
            {'ceph': {
                "sockets": {
                    "/var/run/ceph/osd.asok": {"version": "12.2.0"},
                },
                "installed": True,
                "version": "ceph version 12.2.0 (32ce2a3ae5239ee33d6150705cdb24d43bab910c) luminous (rc)",
            }}
        )
        metadata['mons']['node1'] = node1_data
        result = common.check_ceph_socket_and_installed_version_parity('node1', node1_data)
        assert result is None

    def test_socket_version_is_none(self, make_nodes, make_data):
        metadata['nodes'] = make_nodes(mons=['node1'])
        node1_data = make_data(
            {'ceph': {
                "sockets": {
                    "/var/run/ceph/osd.asok": {},
                },
                "installed": True,
                "version": "12.2.1",
            }}
        )
        metadata['mons']['node1'] = node1_data
        result = common.check_ceph_socket_and_installed_version_parity('node1', node1_data)
        assert result is None



"""
A collection of helpers that will connect to a remote node to run a system
command to return a specific value, instead of shipping a module and executing
functions remotely, this just uses the current connection to execute Popen
"""
import json
from remoto.process import check


def ceph_version(conn):
    try:
        output, _, _ = check(conn, ['ceph', '--version'])
        return output[0]
    except RuntimeError:
        conn.logger.exception('failed to fetch ceph version')


def ceph_socket_version(conn, socket):
    try:
        output, _, _ = check(conn, ['ceph', '--admin-daemon', socket, 'version'])
        result = dict()
        try:
            result = json.loads(output[0])
        except ValueError:
            conn.logger.exception("failed to fetch ceph socket version, invalid json: %s" % output[0])
        return result
    except RuntimeError:
        conn.logger.exception('failed to fetch ceph socket version')


def ceph_is_installed(conn):
    try:
        stdout, stderr, exit_code = check(conn, ['which', 'ceph'])
    except RuntimeError:
        conn.logger.exception('failed to check if ceph is available in the path')
        # XXX this might be incorrect
        return False
    if exit_code != 0:
        return False
    return True



import ceph_medic
from ceph_medic import runner
from ceph_medic.tests import base_metadata


class TestRunner(object):

    def setup(self):
        runner.metadata = base_metadata

    def teardown(self):
        runner.metadata = base_metadata

    def test_calculate_total_hosts_is_0(self):
        run = runner.Runner()
        assert run.total_hosts == 0

    def test_calculate_hosts_single_daemon_type(self):
        ceph_medic.metadata['nodes']['osds'] = [{'host': 'node1'},{'host': 'node2'}]
        runner.metadata = ceph_medic.metadata
        run = runner.Runner()
        assert run.total_hosts == 2

    def test_count_from_different_daemon_types(self):
        ceph_medic.metadata['nodes']['osds'] = [{'host': 'node1'},{'host': 'node2'}]
        ceph_medic.metadata['nodes']['mons'] = [{'host': 'node3'},{'host': 'node4'}]
        runner.metadata = ceph_medic.metadata
        run = runner.Runner()
        assert run.total_hosts == 4


class TestReport(object):

    def setup(self):
        runner.metadata = base_metadata
        runner.metadata['nodes'] = {}
        self.results = runner.Runner()

    def test_reports_errors(self, terminal):
        self.results.errors = ['I am an error']
        runner.report(self.results)
        assert 'While running checks, ceph-medic had unhandled errors' in terminal.calls[-1]

    def test_reports_no_errors(self, terminal):
        runner.report(self.results)
        assert terminal.calls[0] == '\n0 passed, on 0 hosts'


class TestReportBasicOutput(object):

    def setup(self):
        runner.Runner().run()

    def test_has_version(self, terminal):
        assert 'Version: ' in terminal.get_output()

    def test_has_cluster_name(self, terminal):
        assert 'Cluster Name: "ceph"' in terminal.get_output()

    def test_has_no_hosts(self, terminal):
        assert 'Total hosts: [0]' in terminal.get_output()

    def test_has_a_header(self, terminal):
        assert '==  Starting remote check session  ==' in terminal.get_output()

    def test_has_no_OSDs(self, terminal):
        assert 'OSDs:    0' in terminal.get_output()

    def test_has_no_MONs(self, terminal):
        assert 'MONs:    0' in terminal.get_output()

    def test_has_no_Clients(self, terminal):
        assert 'Clients:    0' in terminal.get_output()

    def test_has_no_MDSs(self, terminal):
        assert 'MDSs:    0' in terminal.get_output()

    def test_has_no_MGRs(self, terminal):
        assert 'MGRs:       0' in terminal.get_output()

    def test_has_no_RGWs(self, terminal):
        assert 'RGWs:    0' in terminal.get_output()


class TestReportErrors(object):

    def test_get_new_lines_in_errors(self, terminal, mon_keyring, data, monkeypatch):
        data_node1 = data()
        data_node2 = data()
        data_node1['paths']['/var/lib/ceph']['files'] = {
            '/var/lib/ceph/mon/ceph-0/keyring': {'contents': mon_keyring()}
        }
        data_node1['paths']['/var/lib/ceph']['dirs'] = {
            '/var/lib/ceph/osd/ceph-10': {},
            '/var/lib/ceph/osd/ceph-11': {},
            '/var/lib/ceph/osd/ceph-12': {},
            '/var/lib/ceph/osd/ceph-13': {},
            '/var/lib/ceph/osd/ceph-0': {},
            '/var/lib/ceph/osd/ceph-1': {},
            '/var/lib/ceph/osd/ceph-2': {},
            '/var/lib/ceph/osd/ceph-3': {},
        }

        data_node2['paths']['/var/lib/ceph']['files'] = {
            '/var/lib/ceph/mon/ceph-1/keyring': {'contents': mon_keyring()},
        }
        data_node2['paths']['/var/lib/ceph']['dirs'] = {
            '/var/lib/ceph/osd/ceph-10': {},
            '/var/lib/ceph/osd/ceph-11': {},
            '/var/lib/ceph/osd/ceph-12': {},
            '/var/lib/ceph/osd/ceph-13': {},
            '/var/lib/ceph/osd/ceph-0': {},
            '/var/lib/ceph/osd/ceph-1': {},
            '/var/lib/ceph/osd/ceph-2': {},
            '/var/lib/ceph/osd/ceph-3': {},
        }

        # set the data everywhere we need it
        ceph_medic.metadata['mons'] = {'node1': data_node1, 'node2': data_node2}
        monkeypatch.setattr(ceph_medic.checks.mons, 'metadata', ceph_medic.metadata)

        runner.Runner().run()
        # Any line that is an error or a warning *must* end with a newline
        for line in terminal.calls:
            if line.lstrip().startswith(('E', 'W')):
                assert line.endswith('\n')



config = {
    'verbosity': 'info',
    'nodes': {},
    'hosts_file': None,
}

metadata = {'rgws': {}, 'mgrs': {}, 'mdss':{}, 'clients': {}, 'osds':{}, 'mons':{}, 'nodes': {}}

daemon_types = [i for i in metadata.keys() if i != 'nodes']

__version__ = '1.0.4'



from ceph_medic import metadata
from ceph_medic.util import configuration


#
# Utilities
#

def get_osd_ceph_fsids(data):
    fsids = []
    for file_path in data['paths']['/var/lib/ceph']['files'].keys():
        if "ceph_fsid" in file_path:
            fsids.append(data['paths']['/var/lib/ceph']['files'][file_path]['contents'].strip())
    return set(fsids)


# XXX move out to a central utility module for other checks
def get_ceph_conf(data):
    path = '/etc/ceph/%s.conf' % metadata['cluster_name']
    try:
        conf_file = data['paths']['/etc/ceph']['files'][path]
    except KeyError:
        return None
    return configuration.load_string(conf_file['contents'])


def check_osd_ceph_fsid(host, data):
    code = 'WOSD1'
    msg = "Multiple ceph_fsid values found: %s"

    current_fsids = get_osd_ceph_fsids(data)

    if len(current_fsids) > 1:
        return code, msg % ", ".join(current_fsids)


def check_min_pool_size(host, data):
    code = 'WOSD2'
    msg = 'osd default pool size is set to 1, can potentially lose data'
    conf = get_ceph_conf(data)
    size = conf.get_safe('global', 'osd_pool_default_min_size', '0')
    if int(size) == 1:
        return code, msg


def check_min_osd_nodes(host, data):
    code = 'WOSD3'
    msg = 'OSD nodes might not be enough for a healthy cluster (%s needed, %s found)'
    conf = get_ceph_conf(data)
    default_size = int(conf.get_safe('global', 'osd_pool_default_size', '3'))
    min_size = int(conf.get_safe('global', 'osd_pool_default_min_size', '0'))
    magical_number = default_size + min_size + 1
    osd_nodes = len(metadata['osds'])
    if magical_number > osd_nodes:
        return code, msg % (magical_number, osd_nodes)



"""
JSON Loading utilities
"""
import os
import imp


def load_config(filepath, **kw):
    '''
    Creates a configuration dictionary from a file.

    :param filepath: The path to the file.
    '''

    abspath = os.path.abspath(os.path.expanduser(filepath))
    conf_dict = {}
    if not os.path.isfile(abspath):
        raise RuntimeError('`%s` is not a file.' % abspath)

    # First, make sure the code will actually compile (and has no SyntaxErrors)
    with open(abspath, 'rb') as f:
        compiled = compile(f.read(), abspath, 'exec')

    # Next, attempt to actually import the file as a module.
    # This provides more verbose import-related error reporting than exec()
    absname, _ = os.path.splitext(abspath)
    basepath, module_name = absname.rsplit(os.sep, 1)
    try:
        imp.load_module(
            module_name,
            *imp.find_module(module_name, [basepath])
        )
    except ImportError:
        pass

    # If we were able to import as a module, actually exec the compiled code
    exec(compiled, globals(), conf_dict)
    conf_dict['__file__'] = abspath
    return conf_dict




could_not_connect_to_cluster = """
Failed to connect to cluster when running:

    {command}

This could be because the Ceph cluster name might not map correctly to
a configuration file or a keyring. If using custom cluster names, make sure
that the Ceph configuration file is named the same.

The default cluster name is 'ceph' and ceph-medic will infer the cluster name
if there is no /etc/ceph/ceph.conf by finding the first ".conf" file in that
directory.
"""



import os
import pytest
from ceph_medic.util import configuration
from ceph_medic import log
import logging


class TestLogSetup(object):

    def teardown(self):
        logger = logging.getLogger()
        logger.handlers = []

    def test_barf_when_config_path_does_not_exist(self, tmpdir):
        location = os.path.join(str(tmpdir), 'ceph-medic.conf')
        with open(location, 'w') as _f:
            _f.write("""\n[global]\n--log-path=/bogus/path""")
        config = configuration.load(location)
        with pytest.raises(RuntimeError) as error:
            log.setup(config)
        assert 'value does not exist' in str(error)

    def test_create_log_config_correctly(self, tmpdir):
        tmp_log_path = str(tmpdir)
        location = os.path.join(tmp_log_path, 'ceph-medic.conf')
        with open(location, 'w') as _f:
            _f.write("""\n[global]\n--log-path=%s""" % tmp_log_path)
        config = configuration.load(location)
        log.setup(config)
        logger = logging.getLogger()
        # tox has its own logger now, we need to make sure we are talking about the
        # actual configured ones by ceph-medic
        ceph_medic_loggers = [
            i for i in logger.handlers if 'ceph-medic' in getattr(i, 'baseFilename', '')
        ]
        assert len(ceph_medic_loggers) == 1



"""
Collect remote information on Ceph daemons, store everything in memory and make
it available as a global part of the module so that other checks can consume it
"""
from ceph_medic import metadata, remote, terminal
from ceph_medic.terminal import loader
from ceph_medic.connection import get_connection
from execnet.gateway_bootstrap import HostNotFound
import logging


logger = logging.getLogger(__name__)


def collect_paths(conn):
    """
    Gather all the interesting paths from the remote system, stat them, and
    capture contents when needed.

    Generates a tree path, using the "path of interest" as key, and appending
    the absolute paths of files in the 'files' key and directories in the
    'dirs' key. A small subset of a tree would look
    very similar to::

        {
            '/etc/ceph': {
                'dirs': {
                    '/etc/ceph/ceph.d': {...},
                },
                'files': {
                    '/etc/ceph/ceph.d/ceph.conf': {...},
                },
            }
        }

    Each file and dir in a path tree will contain a set of keys populated
    mostly by calling ``stat`` on the remote system for that absolute path, in
    addition to capturing contents when "interesting files" are dfined. For
    example, the contents of a ``ceph.conf`` file will always be captured. This
    is how that file would look like in a tree path::


        {
            '/etc/ceph/ceph.d/test.conf':
                {
                    'contents': '[osd]\nosd mkfs type = xfs\nosd mkfs options[...]    ',
                    'exception': {},
                    'group': 'ceph',
                    'n_fields': 16,
                    'n_sequence_fields': 10,
                    'n_unnamed_fields': 3,
                    'owner': 'ceph',
                    'st_atime': 1492721509.572292,
                    'st_blksize': 4096,
                    'st_blocks': 8,
                    'st_ctime': 1492721507.880156,
                    'st_dev': 64768L,
                    'st_gid': 167,
                    'st_ino': 100704475,
                    'st_mode': 33188,
                    'st_mtime': 1492721506.1060133,
                    'st_nlink': 1,
                    'st_rdev': 0,
                    'st_size': 650,
                    'st_uid': 167
                },

        }

    .. note:: ``contents`` is captured using ``file.read()`` so its value will
              be a single line with possible line breaks (if any). For reading and
              parsing that key on each line a split must be done on the line break.

    """
    path_metadata = {}
    paths = {
        "/etc/ceph": {'get_contents': True},
        "/var/lib/ceph": {
            'get_contents': True,
            'skip_files': ['activate.monmap', 'superblock'],
            'skip_dirs': ['current', 'store.db']
        },
        "/var/run/ceph": {'get_contents': False},
    }
    for p, kw in paths.items():
        # Collect metadata about the files and dirs for the given path and assign
        # it back to the path_metadata for the current node
        path_metadata[p] = get_path_metadata(conn, p, **kw)
    return path_metadata


def get_path_metadata(conn, path, **kw):
    # generate the tree
    tree = conn.remote_module.path_tree(
        path,
        kw.get('skip_dirs'),
        kw.get('skip_files'),
        kw.get('get_contents')
    )

    files = {}
    dirs = {}

    for i in tree['files']:
        files[i] = conn.remote_module.stat_path(i, None, None, kw.get('get_contents'))
    for i in tree['dirs']:
        dirs[i] = conn.remote_module.stat_path(i, None, None, False)

    # actual root path
    dirs[path] = conn.remote_module.stat_path(path, None, None, False)

    return {'dirs': dirs, 'files': files}


def get_node_metadata(conn, hostname, cluster_nodes):
    # "import" the remote functions so that remote calls using the
    # functions can be executed
    conn.import_module(remote.functions)

    node_metadata = {'ceph': {}}

    # collect paths and files first
    loader.write('Host: %-*s  collecting: [%s]' % (20, hostname, terminal.yellow('paths')))
    node_metadata['paths'] = collect_paths(conn)
    loader.write('Host: %-*s  collecting: [%s]' % (20, hostname, terminal.green('paths')))

    # TODO: collect network information, passing all the cluster_nodes
    # so that it can check for inter-node connectivity
    loader.write('Host: %-*s  collecting: [%s]' % (20, hostname, terminal.yellow('network')))
    node_metadata['network'] = collect_network(cluster_nodes)
    loader.write('Host: %-*s  collecting: [%s]' % (20, hostname, terminal.green('network')))

    # TODO: collect device information
    loader.write('Host: %-*s  collecting: [%s]' % (20, hostname, terminal.yellow('devices')))
    node_metadata['devices'] = collect_devices()
    loader.write('Host: %-*s  collecting: [%s]' % (20, hostname, terminal.green('devices')))

    # collect ceph information
    loader.write('Host: %-*s  collecting: [%s]' % (20, hostname, terminal.yellow('ceph information')))
    node_metadata['ceph'] = collect_ceph_info(conn)
    node_metadata['ceph']['sockets'] = collect_socket_info(conn, node_metadata)
    loader.write('Host: %-*s  collecting: [%s]' % (20, hostname, terminal.green('ceph information')))

    return node_metadata


def collect():
    """
    The main collecting entrypoint. This function will call all the pieces
    needed to build the complete metadata set of a remote system so that checks
    can consume and verify that data.

    After collection is done, the full contents of the metadata are available
    at ``ceph_medic.metadata``
    """
    cluster_nodes = metadata['nodes']
    loader.write('collecting remote node information')
    total_nodes = 0
    failed_nodes = 0
    for node_type, nodes in cluster_nodes.items():
        for node in nodes:
            # check if a node type exists for this node before doing any work:
            try:
                metadata[node_type]
            except KeyError:
                msg = "Skipping node {} from unknown host group: {}".format(node, node_type)
                logger.warning(msg)
                continue

            total_nodes += 1
            hostname = node['host']
            loader.write('Host: %-20s  connection: [%-20s]' % (hostname, terminal.yellow('connecting')))
            # TODO: make sure that the hostname is resolvable, trying to
            # debug SSH issues with execnet is pretty hard/impossible, use
            # util.net.host_is_resolvable
            try:
                logger.debug('attempting connection to host: %s', node['host'])
                conn = get_connection(node['host'])
                loader.write('Host: %-20s  connection: [%-20s]' % (hostname, terminal.green('connected')))
                loader.write('\n')
            except HostNotFound:
                logger.exception('connection failed')
                loader.write('Host: %-20s  connection: [%-20s]' % (hostname, terminal.red('failed')))
                loader.write('\n')
                failed_nodes += 1
                continue

            # send the full node metadata for global scope so that the checks
            # can consume this
            metadata[node_type][node['host']] = get_node_metadata(conn, hostname, cluster_nodes)
            conn.exit()
    if failed_nodes == total_nodes:
        loader.write(terminal.red('Collection failed!') + ' ' *70 + '\n')
        raise RuntimeError('All nodes failed to connect. Cannot run any checks')
    else:
        loader.write('Collection completed!' + ' ' *70 + '\n')


# Network
#
def collect_network(cluster_nodes):
    """
    Collect node-specific information, but also try to check connectivity to
    other hosts that are passed in as ``cluster_nodes``
    """
    return {}


# Devices
#
def collect_devices():
    """
    Get all the device information from the current node
    """
    return {}


# Ceph
#
def collect_ceph_info(conn):
    result = dict()
    result['version'] = remote.commands.ceph_version(conn)
    result['installed'] = remote.commands.ceph_is_installed(conn)
    return result


# Ceph socket info
#
def collect_socket_info(conn, node_metadata):
    sockets = [socket for socket in node_metadata['paths']['/var/run/ceph']['files']
               if socket.endswith(".asok")]
    result = dict()
    for socket in sockets:
        result[socket] = remote.commands.ceph_socket_version(conn, socket)
    return result



from ceph_medic import check, log
import sys
import os
from textwrap import dedent
from tambo import Transport
from execnet.gateway_bootstrap import HostNotFound
import ceph_medic
from ceph_medic.decorators import catches
from ceph_medic.util import configuration
from ceph_medic import terminal


class Medic(object):
    _help = """
ceph-medic: A utility to run system checks on a Ceph cluster.

Version: {version}

Global Options:
  --config              Path to a specific configuration file. Overrides the default:
                        $HOME/.cephmedic.conf.
  --cluster             Use a specific cluster name (defaults to 'ceph'). Alternatively,
                        this is inferred from a conf file name in /etc/ceph/
  --ssh-config          Specify an alternate configuration for SSH
  --version, version    Shows the current installed version
  --inventory           Prefer a ceph-ansible inventory (hosts) file instead of default
                        (cwd, /etc/ansible/hosts) locations

{sub_help}

{config_path_header}: {config_path}
{hosts_file_header}: {hosts_file}
{configured_nodes}
    """
    mapper = {
        'check': check.Check,
        # TODO: this needs a bit more work, disabling for now
        #'generate': generate.Generate,
    }

    def __init__(self, argv=None, parse=True):
        if argv is None:
            argv = sys.argv
        if parse:
            self.main(argv)

    def help(self, sub_help=None):
        if self.hosts_file is None:
            hosts_file_header = terminal.red('Loaded Inventory Hosts file')
            hosts_file = 'No hosts file found in cwd, /etc/ansible/, or configured'
        else:
            hosts_file_header = terminal.green('Loaded Inventory Hosts file')
            hosts_file = self.hosts_file
        return self._help.format(
            version=ceph_medic.__version__,
            config_path=self.config_path,
            config_path_header=terminal.green('Loaded Config Path'),
            hosts_file=hosts_file,
            hosts_file_header=hosts_file_header,
            sub_help=sub_help,
            configured_nodes=self.configured_nodes
        )

    @property
    def configured_nodes(self):
        _help = dedent("""
            Configured nodes (loaded from inventory hosts file):
              OSDs: {osd_node_count}
              MONs: {mon_node_count}
              MGRs: {mgr_node_count}
              MDSs: {mds_node_count}
              RGWs: {rgw_node_count}""")
        if self.hosts_file:  # we have nodes that have been loaded
            nodes = ceph_medic.config['nodes']
            return _help.format(
                osd_node_count=len(nodes.get('osds', [])),
                mon_node_count=len(nodes.get('mons', [])),
                mds_node_count=len(nodes.get('mdss', [])),
                mgr_node_count=len(nodes.get('mgrs', [])),
                rgw_node_count=len(nodes.get('rgws', []))
            )
        return ''

    @catches((RuntimeError, KeyboardInterrupt, HostNotFound))
    def main(self, argv):
        options = [
            '--cluster', '--ssh-config', '--inventory',
            '--config',
        ]
        parser = Transport(
            argv, options=options,
            check_help=False,
            check_version=False
        )
        parser.parse_args()

        self.config_path = parser.get('--config', configuration.location())

        # load medic configuration
        loaded_config = configuration.load(path=parser.get('--config', self.config_path))

        # this is the earliest we can have enough config to setup logging
        log.setup(loaded_config)
        # update the module-wide configuration object
        ceph_medic.config.update(configuration.get_overrides(loaded_config))

        # SSH config
        ceph_medic.config['ssh_config'] = parser.get('--ssh-config')
        if ceph_medic.config['ssh_config']:
            ssh_config_path = ceph_medic.config['ssh_config']
            if not os.path.exists(ssh_config_path):
                terminal.error("the given ssh config path does not exist: %s" % ssh_config_path)
                sys.exit()

        ceph_medic.config['cluster_name'] = parser.get('--cluster')
        ceph_medic.metadata['cluster_name'] = 'ceph'

        # Hosts file
        self.hosts_file = parser.get('--inventory', configuration.get_host_file())

        # find the hosts files, by the CLI first, fallback to the configuration
        # file, and lastly if none of those are found or defined, try to load
        # from well known locations (cwd, and /etc/ansible/)
        loaded_hosts = configuration.load_hosts(
            parser.get('--inventory',
                       ceph_medic.config.get('--inventory', self.hosts_file)))
        ceph_medic.config['nodes'] = loaded_hosts.nodes
        ceph_medic.config['hosts_file'] = loaded_hosts.filename
        self.hosts_file = loaded_hosts.filename

        parser.catch_version = ceph_medic.__version__
        parser.mapper = self.mapper
        parser.catch_help = self.help(parser.subhelp())
        if len(argv) <= 1:
            return parser.print_help()
        ceph_medic.config['config_path'] = self.config_path
        parser.dispatch()
        parser.catches_help()
        parser.catches_version()



import logging
from ceph_medic import metadata, terminal, daemon_types
from ceph_medic import checks, __version__

logger = logging.getLogger(__name__)


class Runner(object):

    def __init__(self):
        self.passed = 0
        self.skipped = 0
        self.failed = 0
        self.total = 0
        self.ignore = []
        self.errors = []

    @property
    def total_hosts(self):
        # XXX does not ensure unique nodes. In collocated scenarios, a single
        # node that is a 'mon' and an 'osd' would count as two nodes
        count = 0
        for daemon in metadata['nodes'].values():
            count += len(daemon)
        return count

    def run(self):
        """
        Go through all the daemons, and all checks. Single entrypoint for running
        checks everywhere.
        """
        start_header()
        for daemon_type in daemon_types:
            self.run_daemons(daemon_type)
        self.total = self.failed + self.passed
        return self

    def run_daemons(self, daemon_type):
        if metadata[daemon_type]:  # we have nodes of this type to run
            nodes_header(daemon_type)

        for host, data in metadata[daemon_type].items():
            modules = [checks.common, getattr(checks, daemon_type, None)]
            self.run_host(host, data, modules)

    def run_host(self, host, data, modules):
        terminal.loader.write(' %s' % terminal.yellow(host))
        has_error = False
        for module in modules:
            checks = collect_checks(module)
            for check in checks:
                try:
                    # TODO: figure out how to skip running a specific check if
                    # the code is ignored, maybe introspecting the function?
                    result = getattr(module, check)(host, data)
                except Exception as error:
                    result = None
                    logger.exception('check had an unhandled error: %s', check)
                    self.errors.append(error)
                if result:
                    code, message = result
                    # XXX This is not ideal, we shouldn't need to get all the way here
                    # to make sure this is actually ignored. (Or maybe it doesn't matter?)
                    if code in self.ignore:
                        self.skipped += 1
                        # avoid writing anything else to the terminal, and just
                        # go to the next check
                        continue
                    self.failed += 1
                    if not has_error:
                        terminal.loader.write(' %s' % terminal.red(host))
                        terminal.write.write('\n')

                    if code.startswith('E'):
                        code = terminal.red(code)
                    elif code.startswith('W'):
                        code = terminal.yellow(code)
                    terminal.write.write("   %s: %s\n" % (code, message))
                    has_error = True
                else:
                    self.passed += 1

        if not has_error:
            terminal.loader.write(' %s\n' % terminal.green(host))


run_errors = terminal.yellow("""
While running checks, ceph-medic had unhandled errors, please look at the
configured log file and report the issue along with the traceback.
""")


def report(results):
    msg = "\n{passed}{failed}{skipped}{errors}{hosts}"

    if results.failed:
        msg = terminal.red(msg)
    elif results.errors:
        msg = terminal.yellow(msg)
    else:
        msg = terminal.green(msg)

    terminal.write.raw(
        msg.format(
            passed="%s passed, " % results.passed,
            failed="%s failed, " % results.failed if results.failed else '',
            skipped="%s skipped, " % results.skipped if results.skipped else '',
            errors="%s errors, " % len(results.errors) if results.errors else '',
            hosts="on %s hosts" % results.total_hosts
        )
    )
    if results.errors:
        terminal.write.raw(run_errors)


start_header_tmpl = """
{title:=^80}
Version: {version: >4}    Cluster Name: "{cluster_name}"
Total hosts: [{total_hosts}]
OSDs: {osds: >4}    MONs: {mons: >4}     Clients: {clients: >4}
MDSs: {mdss: >4}    RGWs: {rgws: >4}     MGRs: {mgrs: >7}
"""


def start_header():
    daemon_totals = dict((daemon, 0) for daemon in daemon_types)
    total_hosts = 0
    for daemon in daemon_types:
        count = len(metadata[daemon].keys())
        total_hosts += count
        daemon_totals[daemon] = count
    terminal.write.raw(start_header_tmpl.format(
        title='  Starting remote check session  ',
        version=__version__,
        total_hosts=total_hosts,
        cluster_name=metadata['cluster_name'],
        **daemon_totals))
    terminal.write.raw('=' * 80)


def nodes_header(daemon_type):
    readable_daemons = {
        'rgws': ' rados gateways ',
        'mgrs': ' managers ',
        'mons': ' mons ',
        'osds': ' osds ',
        'clients': ' clients ',
    }

    terminal.write.bold('\n{daemon:-^30}\n'.format(
        daemon=readable_daemons.get(daemon_type, daemon_type)))


def collect_checks(module):
    checks = [i for i in dir(module) if i.startswith('check')]
    return checks



from mock import Mock
from ceph_medic.remote import commands


class TestCephSocketVersion(object):

    def test_gets_socket_version(self, monkeypatch):
        def mock_check(conn, cmd):
            return (['{"version":"12.2.0"}'], [], [])
        monkeypatch.setattr(commands, 'check', mock_check)
        result = commands.ceph_socket_version(Mock(), '/var/run/ceph/osd.asok')
        assert 'version' in result

    def test_handles_invalid_json(self, monkeypatch):
        def mock_check(conn, cmd):
            return (['version=12.2.0'], [], [])
        monkeypatch.setattr(commands, 'check', mock_check)
        result = commands.ceph_socket_version(Mock(), '/var/run/ceph/osd.asok')
        assert not result



import pytest
import random
from ceph_medic import runner
import ceph_medic
from ceph_medic.tests import base_metadata


class FakeWriter(object):

    def __init__(self):
        self.calls = []
        self.write = self.raw
        self.loader = self

    def raw(self, string):
        self.calls.append(string)

    def bold(self, string):
        self.calls.append(string)

    def get_output(self):
        return '\n'.join(self.calls)


@pytest.fixture(scope='class', autouse=True)
def clear_metadata():
    ceph_medic.metadata = base_metadata


@pytest.fixture
def mon_keyring():
    def make_keyring(default=False):
        if default:
            key = "AQBvaBFZAAAAABAA9VHgwCg3rWn8fMaX8KL01A=="
        else:
            key = "%032x==" % random.getrandbits(128)

        return """
    [mon.]
        key = %s
            caps mon = "allow *"
        """ % key
    return make_keyring


@pytest.fixture
def terminal(monkeypatch):
    fake_writer = FakeWriter()
    monkeypatch.setattr(runner.terminal, 'write', fake_writer)
    return fake_writer


@pytest.fixture
def data():
    """
    Default data structure for remote nodes
    """
    def _data():
        return {
            'ceph': {'installed': True, 'version': '12.2.1'},
            'paths': {
                '/etc/ceph': {'files': {}, 'dirs': {}},
                '/var/lib/ceph': {'files': {}, 'dirs': {}},
            }
        }
    return _data


@pytest.fixture
def make_data(data, **kw):
    """
    Customize basic data structure on remote nodes
    """
    def update(dictionary=None):
        base = data()
        if not dictionary:
            return base
        base.update(dictionary)
        return base
    return update


@pytest.fixture
def make_nodes():
    """
    Helper to generate nodes for daemons
    """
    def make_data(**kw):
        """
        ``kw`` is expected to be a mapping between daemon name and hosts for
        that daemon, like::

            make_data(mons=['node1', 'node2']
        """
        # default set of nodes
        data = dict(
            (k, {}) for k in ['rgws', 'mgrs', 'mdss', 'clients', 'osds', 'mons']
        )
        for daemon, node_names in kw.items():
            data[daemon] = [dict(host=node_name) for node_name in node_names]
        return data
    return make_data



from datetime import datetime
import logging
import os

BASE_FORMAT = "[%(name)s][%(levelname)-6s] %(message)s"
FILE_FORMAT = "[%(asctime)s]" + BASE_FORMAT


def setup(config=None):
    root_logger = logging.getLogger()
    log_path = config.get_safe('global', '--log-path', '.')
    if not os.path.exists(log_path):
        raise RuntimeError('configured ``--log-path`` value does not exist: %s' % log_path)
    date = datetime.strftime(datetime.utcnow(), '%Y-%m-%d')
    log_file = os.path.join(log_path, 'ceph-medic-%s.log' % date)

    root_logger.setLevel(logging.DEBUG)

    # File Logger
    fh = logging.FileHandler(log_file)
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(logging.Formatter(FILE_FORMAT))

    root_logger.addHandler(fh)



from ceph_medic import metadata
from ceph_medic.util import configuration


#
# Utilities
#

def get_fsid(data):
    # FIXME: might want to load this thing into ConfigParser so that we can fetch
    # information. ceph-deploy is a good example on how to do this. See:
    # https://github.com/ceph/ceph-deploy/blob/master/ceph_deploy/conf/ceph.py
    cluster_path = '/etc/ceph/%s.conf' % metadata['cluster_name']
    contents = data['paths']['/etc/ceph']['files'][cluster_path]['contents']
    conf = configuration.load_string(contents)
    try:
        return conf.get_safe('global', 'fsid', '')
    except IndexError:
        return ''

#
# Error checks
#


def check_ceph_conf_exists(host, data):
    cluster_conf = '/etc/ceph/%s.conf' % metadata['cluster_name']

    files = data['paths']['/etc/ceph']['files'].keys()
    if cluster_conf not in files:
        msg = "%s does not exist" % cluster_conf
        return 'ECOM1', msg


def check_ceph_executable_exists(host, data):
    if data['ceph']['installed'] is False:
        return 'ECOM2', 'ceph executable was not found in common paths when running `which`'


def check_var_lib_ceph_dir(host, data):
    code = 'ECOM3'
    exception = data['paths']['/var/lib/ceph']['dirs']['/var/lib/ceph']['exception']
    if exception:
        msg = '/var/lib/ceph could not be parsed: %s' % exception['repr']
        return code, msg


def check_var_lib_ceph_permissions(host, data):
    code = 'ECOM4'
    group = data['paths']['/var/lib/ceph']['dirs']['/var/lib/ceph']['group']
    owner = data['paths']['/var/lib/ceph']['dirs']['/var/lib/ceph']['owner']
    if group == owner != 'ceph':
        msg = '/var/lib/ceph has invalid ownership:  %s:%s, should be ceph:ceph' % (owner, group)
        return code, msg


def check_cluster_fsid(host, data):
    code = 'ECOM5'
    msg = 'fsid "%s" is different than host(s): %s'
    mismatched_hosts = []

    current_fsid = get_fsid(data)

    for daemon, hosts in metadata['nodes'].items():
        for host in hosts:
            hostname = host['host']
            host_fsid = get_fsid(metadata[daemon][hostname])
            if current_fsid != host_fsid:
                mismatched_hosts.append(hostname)

    if mismatched_hosts:
        return code, msg % (current_fsid, ','.join(mismatched_hosts))


def check_ceph_version_parity(host, data):
    code = 'ECOM6'
    msg = '(installed) Ceph version "%s" is different than host(s): %s'
    mismatched_hosts = []
    host_version = data['ceph']['version']
    for daemon, hosts in metadata['nodes'].items():
        for host in hosts:
            hostname = host['host']
            version = metadata[daemon][hostname]['ceph']['version']
            if host_version != version:
                mismatched_hosts.append(hostname)

    if mismatched_hosts:
        return code, msg % (host_version, ','.join(mismatched_hosts))


def check_ceph_socket_and_installed_version_parity(host, data):
    code = 'ECOM7'
    msg = '(installed) Ceph version "%s" is different than version from running socket(s): %s'
    mismatched_sockets = []
    host_version = data['ceph']['version']
    sockets = data['ceph']['sockets']
    for socket, socket_data in sockets.items():
        socket_version = socket_data.get('version')
        if socket_version and socket_version not in host_version:
            mismatched_sockets.append("%s:%s" % (socket, socket_version))

    if mismatched_sockets:
        return code, msg % (host_version, ','.join(mismatched_sockets))



import socket


def host_is_resolvable(hostname, _socket=None):
    _socket = _socket or socket  # just used for testing
    try:
        _socket.getaddrinfo(hostname, 0)
    except _socket.gaierror:
        msg = "hostname: %s is not resolvable" % hostname
        raise RuntimeError(msg)
    return True



import sys


class colorize(str):
    """
    Pretty simple to use::

        colorize.make('foo').bold
        colorize.make('foo').green
        colorize.make('foo').yellow
        colorize.make('foo').red
        colorize.make('foo').blue

    Otherwise you could go the long way (for example if you are
    testing this class)::

        string = colorize('foo')
        string._set_attributes()
        string.red

    """

    def __init__(self, string):
        self.stdout = sys.__stdout__
        self.appends = ''
        self.prepends = ''
        self.isatty = self.stdout.isatty()

    def _set_attributes(self):
        """
        Sets the attributes here because the str class does not
        allow to pass in anything other than a string to the constructor
        so we can't really mess with the other attributes.
        """
        for k, v in self.__colors__.items():
            setattr(self, k, self.make_color(v))

    def make_color(self, color):
        if not self.isatty or self.is_windows:
            return self
        return color + self + '\033[0m' + self.appends

    @property
    def __colors__(self):
        return  dict(
                blue   = '\033[34m',
                green  = '\033[92m',
                yellow = '\033[33m',
                red    = '\033[91m',
                bold   = '\033[1m',
                ends   = '\033[0m'
        )

    @property
    def is_windows(self):
        if sys.platform == 'win32':
            return True
        return False

    @classmethod
    def make(cls, string):
        """
        A helper method to return itself and workaround the fact that
        the str object doesn't allow extra arguments passed in to the
        constructor
        """
        obj = cls(string)
        obj._set_attributes()
        return obj

#
# Common string manipulations
#
red_arrow = colorize.make('-->').red
blue_arrow = colorize.make('-->').blue
yellow = lambda x: colorize.make(x).yellow
blue = lambda x: colorize.make(x).blue
green = lambda x: colorize.make(x).green
red = lambda x: colorize.make(x).red
bold = lambda x: colorize.make(x).bold


CRITICAL = 5
ERROR = 4
WARNING = 3
INFO = 2
DEBUG = 1

_level_names = {
    CRITICAL : 'critical',
    WARNING  : 'warning',
    INFO     : 'info',
    ERROR    : 'error',
    DEBUG    : 'debug'
}

_reverse_level_names = dict((v, k) for (k, v) in _level_names.items())

_level_colors = {
    'remote'   : 'bold',
    'critical' : 'red',
    'warning'  : 'yellow',
    'info'     : 'blue',
    'debug'    : 'blue',
    'error'    : 'red'
}


class _Write(object):

    def __init__(self, _writer=None, prefix='', suffix='', clear_line=False, flush=False):
        self._writer = _writer or sys.stdout
        self.suffix = suffix
        self.prefix = prefix
        self.flush = flush
        self.clear_line = clear_line

    def bold(self, string):
        self.write(bold(string))

    def raw(self, string):
        self.write(string + '\n')

    def write(self, line):
        padding = ''
        if self.clear_line:
            if len(line) > 80:
                padding = ' ' * 10
            else:
                padding = ' ' * (80 - len(line))
        line = line + padding
        self._writer.write(self.prefix + line + self.suffix)
        if self.flush:
            self._writer.flush()


write = _Write()
loader = _Write(prefix='\r', clear_line=True)


class LogMessage(object):

    def __init__(self, level_name, message, writer=None, config_level=None):
        self.level_name = level_name
        self.message = message
        self.writer = writer or sys.stdout
        self.config_level = config_level or self.get_config_level()

    def skip(self):
        if self.level_int >= self.config_level:
            return False
        return True

    def header(self):
        colored = colorize.make(self.base_string)
        return getattr(colored, self.level_color)

    @property
    def base_string(self):
        if self.config_level < 2:
            return "--> [%s]" % self.level_name
        return "-->"

    @property
    def level_int(self):
        if self.level_name == 'remote':
            return 2
        return _reverse_level_names.get(self.level_name, 4)

    @property
    def level_color(self):
        return _level_colors.get(self.level_name, 'info')

    def line(self):
        msg = self.message.rstrip('\n')
        return "%s %s\n" % (self.header(), msg)

    def write(self):
        if not self.skip():
            self.writer.write(self.line())

    def get_config_level(self):
        import ceph_medic
        level = ceph_medic.config.get('verbosity', 'error')
        return _reverse_level_names.get(level, 4)


def error(message):
    return LogMessage('error', message).write()


def debug(message):
    return LogMessage('debug', message).write()


def info(message):
    return LogMessage('info', message).write()


def warning(message):
    return LogMessage('warning', message).write()


def critical(message):
    return LogMessage('critical', message).write()



import pytest

from ceph_medic import collector, metadata
from mock import Mock


class FakeConnRemoteModule(object):
    """
    A fake remote_module class to be used
    with Mocked connection objects.

    This class contains stubbed methods for functions
    in ceph_medic.remote.functions which get their return
    value from the class attribute return_values.

    When creating an instance pass a dictionary that maps
    function names to their return values.
    """

    def __init__(self, return_values):
        self.return_values = return_values

    def stat_path(self, *args, **kwargs):
        return self.return_values.get('stat_path', {})

    def path_tree(self, *args, **kwargs):
        return self.return_values.get('path_tree', {})


def get_tree(files=None, dirs=None):
    if files is None:
        files = ["file1.txt"]
    if dirs is None:
        dirs = ["dir1"]
    tree = dict(
        files=files,
        dirs=dirs,
    )
    return tree


def get_mock_connection(data=None, files=None, dirs=None):
    conn = Mock()
    tree = get_tree(files=files, dirs=dirs)
    default_data = dict(
        path_tree=tree
    )
    data = data or default_data
    conn.remote_module = FakeConnRemoteModule(data)
    return conn


class TestCollectPathMetadata(object):

    def test_metadata_includes_dirs(self):
        conn = get_mock_connection()
        result = collector.get_path_metadata(conn, "/some/path")
        assert "dirs" in result

    def test_metadata_includes_files(self):
        conn = get_mock_connection()
        result = collector.get_path_metadata(conn, "/some/path")
        assert "dirs" in result

    def test_metadata_includes_root_path(self):
        conn = get_mock_connection()
        result = collector.get_path_metadata(conn, "/some/path")
        assert "/some/path" in result["dirs"]

    def test_collects_root_path_when_no_files_or_dirs(self):
        conn = get_mock_connection(files=[], dirs=[])
        result = collector.get_path_metadata(conn, "/some/path")
        assert "/some/path" in result["dirs"]


class TestCollectPaths(object):

    @pytest.mark.parametrize(
        'path',
        ['/etc/ceph', '/var/lib/ceph', '/var/run/ceph'],
    )
    def test_includes_paths(self, path, monkeypatch):
        def mock_metadata(conn, p, **kw):
            return dict()
        monkeypatch.setattr(collector, 'get_path_metadata', mock_metadata)
        result = collector.collect_paths(Mock())
        assert path in result


class TestCollectSocketInfo(object):

    def tests_collects_sockets(self, monkeypatch):
        monkeypatch.setattr(collector.remote.commands, 'ceph_socket_version', lambda conn, socket: dict())
        metadata = {
            'paths': {
                '/var/run/ceph': {'files': ['/var/run/ceph/osd.asok']},
            },
        }
        result = collector.collect_socket_info(Mock(), metadata)
        assert '/var/run/ceph/osd.asok' in result

    def test_ignores_unknown_files(self, monkeypatch):
        monkeypatch.setattr(collector.remote.commands, 'ceph_socket_version', lambda conn, socket: dict())
        metadata = {
            'paths': {
                '/var/run/ceph': {'files': ['/var/run/ceph/osd.asok', '/var/run/ceph/osd.log']},
            },
        }
        result = collector.collect_socket_info(Mock(), metadata)
        assert '/var/run/ceph/osd.log' not in result


class TestCollect(object):

    def test_ignores_unknown_group(self):
        metadata["nodes"] = dict(test=[])
        # raises a RuntimeError because all nodes fail to connect
        with pytest.raises(RuntimeError):
            collector.collect()

    def test_collects_node_metadata(self, monkeypatch):
        metadata["nodes"] = {
            "mons": [{"host": "mon0"}],
            "osds": [{"host": "osd0"}],
        }
        metadata["cluster_name"] = "ceph"
        def mock_metadata(conn, hostname, cluster_nodes):
            return dict(meta="data")
        monkeypatch.setattr(collector, "get_connection", lambda host: Mock())
        monkeypatch.setattr(collector, "get_node_metadata", mock_metadata)
        collector.collect()
        assert "mon0" in metadata["mons"]
        assert "meta" in metadata["mons"]["mon0"]


class TestGetNodeMetadata(object):

    @pytest.mark.parametrize(
        'key',
        ['ceph', 'devices', 'paths', 'network',],
    )
    def test_collects_metadata(self, key, monkeypatch):
        def mock_metadata(*args, **kwargs):
            return dict(meta="data")
        monkeypatch.setattr(collector, "collect_devices", mock_metadata)
        monkeypatch.setattr(collector, "collect_paths", mock_metadata)
        monkeypatch.setattr(collector, "collect_network", mock_metadata)
        monkeypatch.setattr(collector, "collect_ceph_info", mock_metadata)
        monkeypatch.setattr(collector, "collect_socket_info", mock_metadata)
        result = collector.get_node_metadata(Mock(), "mon0", [])
        assert key in result



import remoto
import json
import ceph_medic
from ceph_medic import terminal


def get_mon_report(conn):
    command = [
        'ceph',
        '--cluster=%s' % ceph_medic.metadata['cluster_name'],
        'report'
    ]
    out, err, code = remoto.process.check(
        conn,
        command
    )

    if code > 0:
        terminal.error('failed to connect to the cluster to fetch a report from the monitor')
        terminal.error('command: %s' % ' '.join(command))
        for line in err:
            terminal.error(line)
        raise RuntimeError()

    try:
        return json.loads(b''.join(out).decode('utf-8'))
    except ValueError:
        return {}


def get_cluster_nodes(conn):
    """
    Ask a monitor (with a pre-made connection) about all the nodes in
    a cluster. This will be able to get us all known MONs and OSDs.

    It returns a dictionary with a mapping that looks like::

        {
            'mons': [
                {
                    'host': 'node1',
                    'public_ip': '192.168.1.100',
                },
            ],
            'osds': [
                {
                    'host': 'node2',
                    'public_ip': '192.168.1.101',
                },
                {
                    'host': 'node3',
                    'public_ip': '192.168.1.102',
                },
            ]
        }

    """
    report = get_mon_report(conn)
    nodes = {'mons': [], 'osds': []}
    try:
        # XXX Is this really needed? in what case we wouldn't have a monmap
        # with mons?
        mons = report['monmap']['mons']
    except KeyError:
        raise SystemExit(report)
    for i in mons:
        nodes['mons'].append({
            'host': i['name'],
            'public_ip': _extract_ip_address(i['public_addr'])
        })

    osds = report['osd_metadata']
    for i in osds:
        nodes['osds'].append({
            'host': i['hostname'],
            'public_ip': _extract_ip_address(i['front_addr'])
        })

    return nodes


# XXX does not support IPV6

def _extract_ip_address(string):
    """
    Addresses from Ceph reports can come up with subnets and ports using ':'
    and '/' to identify them properly. Parse those types of strings to extract
    just the IP.
    """
    port_removed = string.split(':')[0]
    return port_removed.split('/')[0]



import os
import subprocess


def which(executable):
    """find the location of an executable"""
    locations = (
        '/usr/local/bin',
        '/bin',
        '/usr/bin',
        '/usr/local/sbin',
        '/usr/sbin',
        '/sbin',
    )

    for location in locations:
        executable_path = os.path.join(location, executable)
        if os.path.exists(executable_path):
            return executable_path


def run(command):
    """
    run a command, return stdout, stderr, and exit code.
    """
    process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True
    )
    stdout = process.stdout.read().splitlines()
    stderr = process.stderr.read().splitlines()
    returncode = process.wait()

    return stdout, stderr, returncode



from ceph_medic import metadata
from ceph_medic.checks import mons


class TestGetSecret(object):

    def setup(self):
        self.data = {
            'paths': {
                '/var/lib/ceph': {
                    'files': {
                        '/var/lib/ceph/mon/ceph-mon-0/keyring': {
                            'contents': '',
                        }
                    }
                }
            }
        }

    def set_contents(self, string, file_path=None):
        file_path = file_path or '/var/lib/ceph/mon/ceph-mon-0/keyring'
        self.data['paths']['/var/lib/ceph']['files'][file_path]['contents'] = string

    def test_get_secret(self):
        contents = """
        [mon.]
               key = AQBvaBFZAAAAABAA9VHgwCg3rWn8fMaX8KL01A==
               caps mon = "allow *"
        """
        self.set_contents(contents)
        result = mons.get_secret(self.data)
        assert result == 'AQBvaBFZAAAAABAA9VHgwCg3rWn8fMaX8KL01A=='

    def test_get_no_secret_empty_file(self):
        result = mons.get_secret(self.data)
        assert result == ''

    def test_get_no_secret_wrong_file(self):
        contents = """
        [mon.]
               caps mon = "allow *"
        """
        self.set_contents(contents)
        result = mons.get_secret(self.data)
        assert result == ''


class TestGetMonitorDirs(object):

    def test_get_monitor_dirs(self):
        result = mons.get_monitor_dirs([
            '/var/lib/ceph/mon/ceph-mon-1',
            '/var/lib/ceph/something'])

        assert result == set(['ceph-mon-1'])

    def test_cannot_get_monitor_dirs(self):
        result = mons.get_monitor_dirs([
            '/var/lib/ceph/osd/ceph-osd-1',
            '/var/lib/ceph/something'])
        assert result == set([])

    def test_get_monitor_dirs_multiple(self):
        result = mons.get_monitor_dirs([
            '/var/lib/ceph/mon/ceph-mon-1',
            '/var/lib/ceph/mon/ceph-mon-3',
            '/var/lib/ceph/mon/ceph-mon-2',
            '/var/lib/ceph/something'])

        assert result == set(['ceph-mon-1', 'ceph-mon-2', 'ceph-mon-3'])

    def test_get_monitor_dirs_nested_multiple(self):
        result = mons.get_monitor_dirs([
            '/var/lib/ceph/mon/ceph-mon-1',
            '/var/lib/ceph/mon/ceph-mon-1/nested/dir/',
            '/var/lib/ceph/mon/ceph-mon-1/other/nested',
            '/var/lib/ceph/mon/ceph-mon-2',
            '/var/lib/ceph/something'])

        assert result == set(['ceph-mon-1', 'ceph-mon-2'])


class TestOsdDirs(object):

    def test_get_osd_dirs_nested_multiple(self):
        result = mons.get_osd_dirs([
            '/var/lib/ceph/osd/ceph-1',
            '/var/lib/ceph/osd/ceph-1/nested/dir/',
            '/var/lib/ceph/osd/ceph-1/other/nested',
            '/var/lib/ceph/osd/ceph-2',
            '/var/lib/ceph/something'])

        assert result == set(['ceph-1', 'ceph-2'])


class TestMonRecommendedCount(object):

    def test_recommended_count_is_met(self, data):
        metadata['mons'] = dict(('mon%s' % count, []) for count in range(6))
        metadata['cluster_name'] = 'ceph'
        osd_data = data()
        result = mons.check_mon_recommended_count(None, osd_data)
        assert result is None

    def test_recommended_count_is_unmet(self, data):
        metadata['mons'] = dict(('mon%s' % count, []) for count in range(1))
        metadata['cluster_name'] = 'ceph'
        osd_data = data()
        code, message = mons.check_mon_recommended_count(None, osd_data)
        assert code == 'WMON3'
        assert message == 'Recommended number of MONs (3) not met: 1'


class TestMonCountIsOdd(object):

    def test_count_is_odd(self, data):
        metadata['mons'] = dict(('mon%s' % count, []) for count in range(3))
        metadata['cluster_name'] = 'ceph'
        osd_data = data()
        result = mons.check_mon_count_is_odd(None, osd_data)
        assert result is None

    def test_recommended_count_is_unmet(self, data):
        metadata['mons'] = dict(('mon%s' % count, []) for count in range(2))
        metadata['cluster_name'] = 'ceph'
        osd_data = data()
        code, message = mons.check_mon_count_is_odd(None, osd_data)
        assert code == 'WMON4'
        assert message == 'Number of MONs is not an odd number: 2'


class TestSingleMon(object):

    def test_is_single(self, data):
        metadata['mons'] = {'mon.0': []}
        metadata['cluster_name'] = 'ceph'
        code, message = mons.check_for_single_mon(None, data())
        assert code == 'WMON5'
        assert message == 'A single monitor was detected: mon.0'

    def test_is_not_single(self, data):
        metadata['mons'] = dict(('mon%s' % count, []) for count in range(2))
        metadata['cluster_name'] = 'ceph'
        result = mons.check_for_single_mon(None, data())
        assert result is None



from textwrap import dedent
from ceph_medic.checks import osds
from ceph_medic import metadata


class TestOSDS(object):

    def test_fails_check_ceph_fsid(self):
        data = {'paths': {'/var/lib/ceph': {'files': {
            '/var/lib/ceph/osd/ceph-0/ceph_fsid': {'contents': "fsid1"},
            '/var/lib/ceph/osd/ceph-1/ceph_fsid': {'contents': "fsid2"},
        }}}}
        result = osds.check_osd_ceph_fsid(None, data)
        assert "WOSD1" in result

    def test_min_pool_size_fails(self, data):
        metadata['cluster_name'] = 'ceph'
        contents = dedent("""
        [global]
        cluster = foo
        osd_pool_default_min_size = 1
        """)
        osd_data = data()
        osd_data['paths']['/etc/ceph']['files']['/etc/ceph/ceph.conf'] = {'contents': contents}
        code, error = osds.check_min_pool_size(None, osd_data)
        assert error == 'osd default pool size is set to 1, can potentially lose data'

    def test_min_pool_size_is_correct(self, data):
        metadata['cluster_name'] = 'ceph'
        contents = dedent("""
        [global]
        cluster = foo
        osd_pool_default_min_size = 2
        """)
        osd_data = data()
        osd_data['paths']['/etc/ceph']['files']['/etc/ceph/ceph.conf'] = {'contents': contents}
        result = osds.check_min_pool_size(None, osd_data)
        assert result is None


class TestMinOSDS(object):

    def test_min_osd_nodes_is_not_met(self, data):
        metadata['osds'] = {'osd1': []}
        metadata['cluster_name'] = 'ceph'
        osd_data = data()
        contents = dedent("""
        [global]
        cluster = foo
        osd_pool_default_min_size = 2
        """)
        osd_data['paths']['/etc/ceph']['files']['/etc/ceph/ceph.conf'] = {'contents': contents}
        code, error = osds.check_min_osd_nodes(None, osd_data)
        assert code == 'WOSD3'
        assert '6 needed, 1 found' in error

    def test_min_osd_nodes_is_met(self, data):
        metadata['osds'] = dict(('osd%s' % count, []) for count in range(6))
        metadata['cluster_name'] = 'ceph'
        osd_data = data()
        contents = dedent("""
        [global]
        cluster = foo
        osd_pool_default_min_size = 2
        """)
        osd_data['paths']['/etc/ceph']['files']['/etc/ceph/ceph.conf'] = {'contents': contents}
        result = osds.check_min_osd_nodes(None, osd_data)
        assert result is None



import logging
import socket
import remoto
import ceph_medic

logger = logging.getLogger(__name__)


def get_connection(hostname, username=None, threads=5, use_sudo=None, detect_sudo=True, **kw):
    """
    A very simple helper, meant to return a connection
    that will know about the need to use sudo.
    """
    if kw.get('logger') is False:  # explicitly disable remote logging
        remote_logger = None
    else:
        remote_logger = logging.getLogger(hostname)

    if username:
        hostname = "%s@%s" % (username, hostname)

    if ceph_medic.config.get('ssh_config'):
        hostname = "-F %s %s" % (ceph_medic.config.get('ssh_config'), hostname)
    try:

        conn = remoto.Connection(
            hostname,
            logger=remote_logger,
            threads=threads,
            detect_sudo=detect_sudo,
        )

        # Set a timeout value in seconds to disconnect and move on
        # if no data is sent back.
        conn.global_timeout = 300

        # XXX put this somewhere else
        if not ceph_medic.config['cluster_name']:
            cluster_conf_files, stderr, exit_code = remoto.process.check(conn, ['ls', '/etc/ceph/'])
            cluster_name = 'ceph'
            if 'ceph.conf' not in cluster_conf_files:
                logger.warning('/etc/ceph/ceph.conf was not found, will try to infer the cluster name')
                for i in cluster_conf_files:
                    if i.endswith('conf'):
                        cluster_name = i.split('.conf')[0]
                        logger.warning('inferred %s as the cluster name', cluster_name)
            ceph_medic.metadata['cluster_name'] = cluster_name
        else:
            ceph_medic.metadata['cluster_name'] = ceph_medic.config['cluster_name']
        return conn
    except Exception as error:
        msg = "connecting to host: %s " % hostname
        errors = "resulted in errors: %s %s" % (error.__class__.__name__, error)
        logger.error(msg)
        logger.error(errors)
        raise error


def get_local_connection(logger, use_sudo=False):
    """
    Helper for local connections that are sometimes needed to operate
    on local hosts
    """
    return get_connection(
        socket.gethostname(),  # cannot rely on 'localhost' here
        None,
        logger=logger,
        threads=1,
        use_sudo=use_sudo,
        detect_sudo=False
    )



import pytest
import ceph_medic.main


class TestMain(object):
    def test_main(self):
        assert ceph_medic.main

    def test_invalid_ssh_config(self):
        argv = ["ceph-medic", "--ssh-config", "/does/not/exist"]
        with pytest.raises(SystemExit):
            ceph_medic.main.Medic(argv)



import os

from ceph_medic.remote import functions


def make_test_file(filename, contents=None):
    contents = contents or "foo"
    with open(filename, 'w') as f:
        f.write(contents)


def make_test_tree(path, contents=None, tree=None):
    file1 = os.path.join(path, "file1.txt")
    dir1 = os.path.join(path, "dir1")
    file2 = os.path.join(path, "dir1/file2.txt")
    make_test_file(file1)
    os.mkdir(dir1)
    make_test_file(file2)


class TestStatPath(object):

    def test_stat_file_includes_owner(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'file')
        make_test_file(filename)

        result = functions.stat_path(filename)
        assert "owner" in result

    def test_stat_file_includes_group(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'file')
        make_test_file(filename)

        result = functions.stat_path(filename)
        assert "group" in result

    def test_includes_file_content(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'file')
        make_test_file(filename, contents="foo")

        result = functions.stat_path(filename, get_contents=True)
        assert result["contents"] == "foo"

    def test_exception_is_empty_on_success(self, tmpdir):
        filename = os.path.join(str(tmpdir), 'file')
        make_test_file(filename)

        result = functions.stat_path(filename)
        assert not result["exception"]

    def test_stat_dir(self, tmpdir):
        result = functions.stat_path(str(tmpdir))
        assert result


class TestPathTree(object):

    def test_skip_dirs(self, tmpdir):
        path = str(tmpdir)
        make_test_tree(path)
        result = functions.path_tree(path, skip_dirs=['dir1'])
        assert "dir1" not in result["dirs"]

    def test_skip_files(self, tmpdir):
        path = str(tmpdir)
        make_test_tree(path)
        result = functions.path_tree(path, skip_files=['file1.txt'])
        assert "file1.txt" not in result["files"]

    def test_includes_path(self, tmpdir):
        path = str(tmpdir)
        make_test_tree(path)
        result = functions.path_tree(path)
        assert result["path"] == path

    def test_includes_files(self, tmpdir):
        path = str(tmpdir)
        make_test_tree(path)
        result = functions.path_tree(path)
        assert "files" in result
        assert os.path.join(path, "file1.txt") in result["files"]

    def test_includes_dirs(self, tmpdir):
        path = str(tmpdir)
        make_test_tree(path)
        result = functions.path_tree(path)
        assert "dirs" in result
        assert os.path.join(path, "dir1") in result["dirs"]



import os
import grp
import pwd
import traceback
import sys
import subprocess


# Utilities
#
def capture_exception(error):
    details = {'attributes': {}}
    details['name'] = error.__class__.__name__
    details['repr'] = str(error)
    exc_type, exc_value, exc_traceback = sys.exc_info()
    details['traceback'] = ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))
    for attr in dir(error):
        if not attr.startswith('__'):
            details['attributes'][attr] = getattr(error, attr)
    return details


# Paths
#
def stat_path(path, skip_dirs=None, skip_files=None, get_contents=False):
    """stat a path on a remote host"""
    # Capture all information about a path, optionally getting the contents of
    # the remote path if it is a file. Exceptions get appended to each dictionary
    # object associated with the path

    # .. note:: Neither ``skip_dirs`` nor ``skip_files`` is used here, but the
    # remote execution of functions use name-based arguments which does not allow
    # the use of ``**kw``

    metadata = {'exception': {}}
    try:
        stat_info = os.stat(path)
        if get_contents and os.path.isfile(path):
            with open(path, 'r') as opened_file:
                metadata['contents'] = opened_file.read()
    except Exception as error:
        return {'exception': capture_exception(error)}

    # get all the stat results back into the metadata
    for attr in dir(stat_info):
        if not attr.startswith('__'):
            metadata[attr] = getattr(stat_info, attr)

    # translate the owner and group:
    metadata['owner'] = pwd.getpwuid(stat_info.st_uid)[0]
    metadata['group'] = grp.getgrgid(stat_info.st_gid)[0]

    return metadata


def path_tree(path, skip_dirs=None, skip_files=None, get_contents=None):
    """generate a path tree"""
    # Generate a tree of paths, including directories and files, recursively, but
    # with the ability to exclude dirs and files with ``skip_dirs`` and
    # ``skip_files``.
    # The tree output groups the files and directories like::

    #     {
    #         'path': '/etc/ceph',
    #         'dirs': ['/etc/ceph/ceph.d/'],
    #         'files': ['/etc/ceph/ceph.d/test.conf', '/etc/ceph/rbdmap']
    #     }

    # .. note:: ``get_contents`` is not used here, but the remote execution of functions
    # use name-based arguments which does not allow the use of ``**kw``

    skip_files = skip_files or []
    skip_dirs = skip_dirs or []
    files = []
    dirs = []
    # traverse for files and directories, topdown allows us to trim the
    # directories on the fly
    for root, _dirs, _files in os.walk(path, topdown=True):
        _dirs[:] = [d for d in _dirs if d not in skip_dirs]
        for _file in _files:
            absolute_path = os.path.join(root, _file)
            if _file in skip_files:
                continue
            files.append(absolute_path)

        for _dir in _dirs:
            absolute_path = os.path.join(root, _dir)
            dirs.append(absolute_path)

    return dict(path=path, dirs=dirs, files=files)


def which(executable):
    """find the location of an executable"""
    locations = (
        '/usr/local/bin',
        '/bin',
        '/usr/bin',
        '/usr/local/sbin',
        '/usr/sbin',
        '/sbin',
    )

    for location in locations:
        executable_path = os.path.join(location, executable)
        if os.path.exists(executable_path):
            return executable_path


def run(command):
    """
    run a command, return stdout, stderr, and exit code.
    """
    process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True
    )
    stdout = process.stdout.read().splitlines()
    stderr = process.stderr.read().splitlines()
    returncode = process.wait()

    return stdout, stderr, returncode


# remoto magic, needed to execute these functions remotely
if __name__ == '__channelexec__':
    for item in channel:  # noqa
        channel.send(eval(item))  # noqa



import sys
import ceph_medic
import logging
from ceph_medic import runner, collector
from tambo import Transport

logger = logging.getLogger(__name__)


class Check(object):
    help = "Run checks for all the configured nodes in a cluster or hosts file"
    long_help = """
check: Run for all the configured nodes in the configuration

Options:
  --ignore              Comma-separated list of errors and warnings to ignore.


Loaded Config Path: {config_path}

Configured Nodes:
{configured_nodes}
    """

    def __init__(self, argv=None, parse=True):
        self.argv = argv or sys.argv

    @property
    def subcommand_args(self):
        # find where `check` is
        index = self.argv.index('check')
        # slice the args
        return self.argv[index:]

    def _help(self):
        node_section = []
        for daemon, node in ceph_medic.config['nodes'].items():
            header = "\n* %s:\n" % daemon
            body = '\n'.join(["    %s" % n for n in ceph_medic.config['nodes'][daemon]])
            node_section.append(header+body+'\n')
        return self.long_help.format(
            configured_nodes=''.join(node_section),
            config_path=ceph_medic.config['config_path']
        )

    def main(self):
        options = ['--ignore']
        parser = Transport(
            self.argv, options=options,
            check_version=False
        )
        parser.catch_help = self._help()

        parser.parse_args()
        if len(self.argv) < 1:
            return parser.print_help()

        # populate the nodes metadata with the configured nodes
        for daemon in ceph_medic.config['nodes'].keys():
            ceph_medic.metadata['nodes'][daemon] = []
        for daemon, nodes in ceph_medic.config['nodes'].items():
            for node in nodes:
                ceph_medic.metadata['nodes'][daemon].append({'host': node['host']})

        collector.collect()
        test = runner.Runner()
        results = test.run()
        runner.report(results)



from ceph_medic import terminal


class FakeWriter(object):

    def __init__(self):
        self.calls = []

    def write(self, string):
        self.calls.append(string)

    def flush(self):
        pass


class TestWriteClearLine(object):

    def setup(self):
        self.fake_writer = FakeWriter()
        self.loader = terminal._Write(
            _writer=self.fake_writer,
            prefix='\r',
            clear_line=True
        )

    def test_adds_padding_for_81_chars(self):
        self.loader.write('1234567890')
        assert len(self.fake_writer.calls[0]) == 81

    def test_remaining_padding_is_whitespace(self):
        self.loader.write('1234567890')
        assert self.fake_writer.calls[0][11:] == ' ' * 70

    def test_long_line_adds_only_ten_chars(self):
        self.loader.write('1'*81)
        assert self.fake_writer.calls[0][82:] == ' ' * 10



from setuptools import setup, find_packages, Command
import re
import sys
import subprocess


install_requires = []
pyversion = sys.version_info[:2]


def read_module_contents():
    with open('ceph_medic/__init__.py') as f:
        return f.read()


module_file = read_module_contents()
metadata = dict(re.findall("__([a-z]+)__\s*=\s*'([^']+)'", module_file))
long_description = open('README.rst').read()
version = metadata['version']


class BumpCommand(Command):
    """ Bump the __version__ number and commit all changes. """

    user_options = [('version=', 'v', 'version number to use')]

    def initialize_options(self):
        new_version = metadata['version'].split('.')
        new_version[-1] = str(int(new_version[-1]) + 1)  # Bump the final part
        self.version = ".".join(new_version)

    def finalize_options(self):
        pass

    def run(self):

        try:
            print('old version: %s  new version: %s' %
                  (metadata['version'], self.version))
            raw_input('Press enter to confirm, or ctrl-c to exit >')
        except KeyboardInterrupt:
            raise SystemExit("\nNot proceeding")

        old = "__version__ = '%s'" % metadata['version']
        new = "__version__ = '%s'" % self.version

        module_file = read_module_contents()
        with open('ceph_medic/__init__.py', 'w') as fileh:
            fileh.write(module_file.replace(old, new))

        # Commit everything with a standard commit message
        cmd = ['git', 'commit', '-a', '-m', 'version %s' % self.version]
        print(' '.join(cmd))
        subprocess.check_call(cmd)


class ReleaseCommand(Command):
    """ Tag and push a new release. """

    user_options = [('sign', 's', 'GPG-sign the Git tag and release files')]

    def initialize_options(self):
        self.sign = False

    def finalize_options(self):
        pass

    def run(self):
        # Create Git tag
        tag_name = 'v%s' % version
        cmd = ['git', 'tag', '-a', tag_name, '-m', 'version %s' % version]
        if self.sign:
            cmd.append('-s')
        print(' '.join(cmd))
        subprocess.check_call(cmd)

        # Push Git tag to origin remote
        cmd = ['git', 'push', 'origin', tag_name]
        print(' '.join(cmd))
        subprocess.check_call(cmd)

        # Push package to pypi
        cmd = ['python', 'setup.py', 'sdist', 'upload']
        if self.sign:
            cmd.append('--sign')
        print(' '.join(cmd))
        #subprocess.check_call(cmd)

        # Push master to the remote
        cmd = ['git', 'push', 'origin', 'master']
        print(' '.join(cmd))
        subprocess.check_call(cmd)

setup(
    name='ceph-medic',
    version=version,
    packages=find_packages(),

    author='Alfredo Deza',
    author_email='contact@redhat.com',
    description='detect common issues with ceph clusters',
    long_description=long_description,
    license='MIT',
    keywords='ceph doctor',
    url="https://github.com/ceph/ceph-medic",
    zip_safe=False,

    install_requires=[
        'execnet',
        'tambo',
        'remoto',
    ] + install_requires,

    tests_require=[
        'pytest >=2.1.3',
        'tox',
        'mock',
    ],

    scripts=['bin/ceph-medic'],
    cmdclass={'bump': BumpCommand, 'release': ReleaseCommand},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Utilities',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
    ]

)



from ceph_medic import metadata
from ceph_medic.util import configuration

#
# Utilities
#


def get_secret(data):
    """
    keyring files look like::

    [mon.]
        key = AQBvaBFZAAAAABAA9VHgwCg3rWn8fMaX8KL01A==
            caps mon = "allow *"

    Fetch that keyring file and extract the actual key, no spaces.

    .. warning:: If multiple mon dirs exist, this utility will pick the first
    one it finds. There are checks that will complain about multiple mon dirs
    """
    file_paths = data['paths']['/var/lib/ceph']['files'].keys()
    _path = data['paths']['/var/lib/ceph']['files']
    for _file in file_paths:
        if _file.startswith('/var/lib/ceph/mon/') and _file.endswith('keyring'):
            contents = _path[_file]['contents']
            conf = configuration.load_string(contents)
            try:
                return conf.get_safe('mon.', 'key', '').split('\n')[0]
            except IndexError:
                # is it really possible to get a keyring file that doesn't
                # have a monitor secret?
                return ''


def get_monitor_dirs(dirs):
    """
    Find all the /var/lib/ceph/mon/* directories. This is a bit tricky because
    we don't know if there are nested directories (the metadata reports them in
    a flat list).
    We must go through all of them and make sure that by splitting there aren't
    any nested ones and we are only reporting actual monitor dirs.
    """
    # get all the actual monitor dirs
    found = []
    prefix = '/var/lib/ceph/mon/'
    mon_dirs = [d for d in dirs if d.startswith(prefix)]
    for _dir in mon_dirs:
        # splitting on prefix[-1] will give us:
        # 'ceph-mon-1/maybe/nested' or 'ceph-mon-1'
        dirs = _dir.split(prefix)[-1].split('/')
        # splitting again on '/' and using the first part will ensure we only
        # get the dir
        found.append(dirs[0])
    return set(found)


def get_osd_dirs(dirs):
    """
    Find all the /var/lib/ceph/osd/* directories. This is a bit tricky because
    we don't know if there are nested directories (the metadata reports them in
    a flat list).
    We must go through all of them and make sure that by splitting there aren't
    any nested ones and we are only reporting actual monitor dirs.
    """
    # get all the actual monitor dirs
    found = []
    prefix = '/var/lib/ceph/osd/'
    osd_dirs = [d for d in dirs if d.startswith(prefix)]
    for _dir in osd_dirs:
        # splitting on prefix[-1] will give us:
        # 'ceph-1/maybe/nested' or 'ceph-1'
        dirs = _dir.split(prefix)[-1].split('/')
        # splitting again on '/' and using the first part will ensure we only
        # get the dir
        found.append(dirs[0])
    return set(found)
#
# Error Checks
#


def check_mon_secret(host, data):
    code = 'EMON1'
    msg = 'secret key "%s" is different than host(s): %s'
    mismatched_hosts = []

    current_secret = get_secret(data)
    if not current_secret:
        # there is no file for the current host, so we can't compare
        return

    for host, host_data in metadata['mons'].items():
        host_secret = get_secret(host_data)
        if not host_secret:
            # cannot compare with another host that may not have the secret
            continue
        if current_secret != host_secret:
            mismatched_hosts.append(host)

    if mismatched_hosts:
        return code, msg % (current_secret, ','.join(mismatched_hosts))

#
# Warning Checks
#


def check_multiple_mon_dirs(host, data):
    code = 'WMON1'
    msg = 'multiple /var/lib/ceph/mon/* dirs found: %s'
    dirs = data['paths']['/var/lib/ceph']['dirs']
    monitor_dirs = get_monitor_dirs(dirs)
    if len(monitor_dirs) > 1:
        return code, msg % ','.join(monitor_dirs)


def check_mon_collocated_with_osd(host, data):
    code = 'WMON2'
    msg = 'collocated OSDs found: %s'
    dirs = data['paths']['/var/lib/ceph']['dirs']
    osd_dirs = get_osd_dirs(dirs)
    if len(osd_dirs):
        return code, msg % ','.join(osd_dirs)


def check_mon_recommended_count(host, data):
    code = 'WMON3'
    msg = 'Recommended number of MONs (3) not met: %s'
    mon_count = len(metadata['mons'].keys())
    if mon_count < 3:
        return code, msg % mon_count


def check_mon_count_is_odd(host, data):
    code = 'WMON4'
    msg = 'Number of MONs is not an odd number: %s'
    mon_count = len(metadata['mons'].keys())
    if mon_count % 2 == 0:
        return code, msg % mon_count


def check_for_single_mon(host, data):
    code = 'WMON5'
    msg = 'A single monitor was detected: %s'
    monitors = list(metadata['mons'].keys())
    if len(monitors) == 1:
        return code, msg % monitors.pop()



# Rules to apply for Jewel releases.

# All checks are applied, but overrides to defaults can
# be specified here.
# overrides = {
#     # overrides the check called 'check_name' with a different expected value
#     # and changes # the level of this check to 'error'.
#     "check_name": {"expected": "value", "level": "error"},
#}

# Exclude the following checks:
# excludes = ["check_name"]

# Include the following checks:
# includes = ["check_name"]
